# -*- coding: utf-8 -*-

import os, sys
import numpy as np
from skimage import color

__all__ = ["mean_patch_val",
           "std_patch_val",
           "patch_bk_ratio"]


def mean_patch_val(img):
    """ Mean pixel value of the patch.

    Parameters
    -------
    img: np.array
        Patch image

    Returns
    -------
    mean_val: float
        Mean pixel value of the patch

    """

    mean_val = np.mean(img)

    return mean_val


def std_patch_val(img):
    """ Standard deviation of pixel values in the patch.

    Parameters
    -------
    img: np.array
        Patch image

    Returns
    -------
    std_val: float
        Standard deviation of pixel values in the patch

    """

    std_val = np.std(img)

    return std_val


def patch_bk_ratio(img, bk_thresh=0.80):
    """ Calculate the ratio of background in the image

    Parameters
    -------
    img: np.array
        patch image
    bk_thresh: float
        background threshold value

    Returns
    -------
    bk_ratio: float
        the ratio of background in a patch

    """

    g_img = color.rgb2gray(img)
    bk_num = (g_img > bk_thresh).sum()
    pixel_num = g_img.shape[0] * g_img.shape[1]
    bk_ratio = bk_num * 1.0 / pixel_num

    return bk_ratio



# -*- coding: utf-8 -*-

import numpy as np
from ._rela import cnt_inside_ratio

__all__ = ["contour_patch_splitting_no_overlap",
           "contour_patch_splitting_self_overlap",
           "contour_patch_splitting_half_overlap",]


def contour_patch_splitting_no_overlap(cnt_arr, wsi_h, wsi_w,
                                       patch_size=299, inside_ratio=0.75):
    """ Splitting contour into patches with no overlapping between patches.

    Parameters
    -------
    cnt_arr: np.array
        contour with standard numpy 2d array format
    wsi_h: int
        the height of whole slide image
    wsi_w: int
        the width of whole slide image
    patch_size: int
        size of patch
    inside_ratio: float
        the ratio of patch to be inside the contour

    Returns
    -------
    coors_arr: list
        list of starting coordinates of patches ([0]-h, [1]-w)

    """

    cnt_min_h, cnt_min_w = np.min(cnt_arr[0, :]), np.min(cnt_arr[1, :])
    cnt_max_h, cnt_max_w = np.max(cnt_arr[0, :]), np.max(cnt_arr[1, :])
    if cnt_min_h < 0 or cnt_min_w < 0 or cnt_max_h > wsi_h or cnt_max_w > wsi_w:
        return []

    # add border to top left
    start_h, start_w = None, None
    half_patch_size = int(np.floor(patch_size / 2.0))
    quarter_patch_size = int(np.floor(patch_size / 4.0))
    if cnt_min_h >= half_patch_size:
        start_h = cnt_min_h - half_patch_size
    elif cnt_min_h >= quarter_patch_size:
        start_h = cnt_min_h - quarter_patch_size
    else:
        start_h = cnt_min_h
    if cnt_min_w >= half_patch_size:
        start_w = cnt_min_w - half_patch_size
    elif cnt_min_w >= quarter_patch_size:
        start_w = cnt_min_w - quarter_patch_size
    else:
        start_w = cnt_min_w

    # make up the border to satisfy patch grids
    end_h = (1 + int(np.floor((cnt_max_h - start_h - 1.0) / patch_size))) * patch_size + start_h
    if end_h > wsi_h:
        end_h -= patch_size
    end_w = (1 + int(np.floor((cnt_max_w - start_w  - 1.0) / patch_size))) * patch_size + start_w
    if end_w > wsi_w:
        end_w -= patch_size

    coors_arr = []
    for cur_h in np.arange(start_h, end_h, patch_size):
        for cur_w in np.arange(start_w, end_w, patch_size):
            cur_patch_cnt = np.array([[cur_h, cur_h, cur_h+patch_size, cur_h+patch_size],
                                      [cur_w, cur_w+patch_size, cur_w+patch_size, cur_w]])
            # inside ratio should sastify conditions to be used
            if cnt_inside_ratio(cur_patch_cnt, cnt_arr) >= inside_ratio:
                coors_arr.append([cur_h, cur_w, patch_size, patch_size])

    return coors_arr


def contour_patch_splitting_self_overlap(cnt_arr, patch_size=299, inside_ratio=0.75):
    """ Splitting contour into patches with both start and end meeting,
    with overlapping among patches.

    Parameters
    -------
    cnt_arr: np.array
        contour with standard numpy 2d array format
    patch_size: int
        size of patch
    inside_ratio: float
        the ratio of patch to be inside the contour

    Returns
    -------
    coors_arr: list
        list of starting coordinates of patches ([0]-h, [1]-w)

    """

    cnt_min_h, cnt_min_w = np.min(cnt_arr[0, :]), np.min(cnt_arr[1, :])
    cnt_max_h, cnt_max_w = np.max(cnt_arr[0, :]), np.max(cnt_arr[1, :])

    cnt_h, cnt_w = cnt_max_h - cnt_min_h, cnt_max_w - cnt_min_w
    h_points = int(np.ceil(cnt_h * 1.0 / patch_size))
    w_points = int(np.ceil(cnt_w * 1.0 / patch_size))

    overlap_h_len = (h_points * patch_size - cnt_h) * 1.0 / (h_points - 1)
    extend_h_len = patch_size - overlap_h_len
    overlap_w_len = (w_points * patch_size - cnt_w) * 1.0 / (w_points - 1)
    extend_w_len = patch_size - overlap_w_len

    coors_arr = []
    for h_ind in np.arange(h_points):
        for w_ind in np.arange(w_points):
            cur_h = int(round(cnt_min_h + extend_h_len * h_ind))
            cur_w = int(round(cnt_min_w + extend_w_len * w_ind))
            cur_patch_cnt = np.array([[cur_h, cur_h, cur_h+patch_size, cur_h+patch_size],
                                      [cur_w, cur_w+patch_size, cur_w+patch_size, cur_w]])
            if cnt_inside_ratio(cur_patch_cnt, cnt_arr) >= inside_ratio:
                coors_arr.append([cur_h, cur_w, patch_size, patch_size])
    return coors_arr


def contour_patch_splitting_half_overlap(cnt_arr, wsi_h, wsi_w,
                                       patch_size=448, inside_ratio=0.75):
    """ Splitting patches with half overlap between patches.

    Parameters
    -------
    cnt_arr: np.array
        contour with standard numpy 2d array format
    wsi_h: int
        the height of whole slide image
    wsi_w: int
        the width of whole slide image
    patch_size: int
        size of patch
    inside_ratio: float
        the ratio of patch to be inside the contour

    Returns
    -------
    coors_arr: list
        list of starting coordinates of patches ([0]-h, [1]-w)

    """

    cnt_min_h, cnt_min_w = np.min(cnt_arr[0, :]), np.min(cnt_arr[1, :])
    cnt_max_h, cnt_max_w = np.max(cnt_arr[0, :]), np.max(cnt_arr[1, :])
    if cnt_min_h < 0 or cnt_min_w < 0 or cnt_max_h > wsi_h or cnt_max_w > wsi_w:
        return []

    # add border to top left
    start_h, start_w = None, None
    half_patch_size = int(np.floor(patch_size / 2.0))
    quarter_patch_size = int(np.floor(patch_size / 4.0))
    if cnt_min_h >= half_patch_size:
        start_h = cnt_min_h - half_patch_size
    elif cnt_min_h >= quarter_patch_size:
        start_h = cnt_min_h - quarter_patch_size
    else:
        start_h = cnt_min_h
    if cnt_min_w >= half_patch_size:
        start_w = cnt_min_w - half_patch_size
    elif cnt_min_w >= quarter_patch_size:
        start_w = cnt_min_w - quarter_patch_size
    else:
        start_w = cnt_min_w

    # make up the border to satisfy patch grids
    end_h = (1 + int(np.floor((cnt_max_h - start_h - 1.0) / half_patch_size))) * half_patch_size + start_h
    if end_h > wsi_h - patch_size:
        end_h -= patch_size
    end_w = (1 + int(np.floor((cnt_max_w - start_w  - 1.0) / half_patch_size))) * half_patch_size + start_w
    if end_w > wsi_w - patch_size:
        end_w -= patch_size

    coors_arr = []
    for cur_h in np.arange(start_h, end_h+half_patch_size, half_patch_size):
        for cur_w in np.arange(start_w, end_w+half_patch_size, half_patch_size):
            cur_patch_cnt = np.array([[cur_h, cur_h, cur_h+patch_size, cur_h+patch_size],
                                      [cur_w, cur_w+patch_size, cur_w+patch_size, cur_w]])
            # inside ratio should sastify conditions to be used
            if cnt_inside_ratio(cur_patch_cnt, cnt_arr) >= inside_ratio:
                coors_arr.append([cur_h, cur_w, patch_size, patch_size])

    return coors_arr



# -*- coding: utf-8 -*-


from pycontour.poly_transform import np_arr_to_poly, poly_to_np_arr

__all__ = ["contour_to_poly_valid",
           ]


def contour_to_poly_valid(cnt_arr):
    """ Convert contour to poly valid if not poly valid

    Parameters
    -------
    cnt_arr: np.array
        contour with standard numpy 2d array format

    Returns
    -------
    cnt_valid_arr: np.array
        contour with standard numpy 2d array format

    """

    poly = np_arr_to_poly(cnt_arr)
    if poly.is_valid == True:
        cnt_valid_arr = cnt_arr
    else:
        cnt_valid_arr = poly_to_np_arr(poly.convex_hull)

    return cnt_valid_arr



# -*- coding: utf-8 -*-

from pycontour.poly_transform import np_arr_to_poly


__all__ = ["contour_valid",
           ]


def contour_valid(cnt_arr):
    """ Check contour is valid or not.

    Parameters
    -------
    cnt_arr: np.array
        contour with standard numpy 2d array format

    Returns
    -------
    valid: boolean
        True if valid, else False

    """

    poly = np_arr_to_poly(cnt_arr)
    valid = True if poly.is_valid else False

    return valid



# -*- coding: utf-8 -*-

import numpy as np
from shapely import geometry

__all__ = ["cnt_inside_wsi",
           "intersect_cnt_wsi",
           "cnt_inside_ratio",
           ]


def cnt_inside_wsi(cnt_arr, wsi_h, wsi_w):
    """ Determine contour is fully inside whole slide image or not.

    Parameters
    -------
    cnt_arr: np.array
        contour with standard numpy 2d array format
    wsi_h: int
        height of whole slide image
    wsi_w: int
        width of whole slide image

    Returns
    -------
    in_flag: bool
        in_flag to be true if contour is fully inside whole slide image,
        else false

    """

    # construct whole slide image polygon, in whole slide image, we need to avoid
    # contour on the maximum width and height line, thus substract a small value
    wsi_poly = geometry.box(0, 0, wsi_w - 0.001, wsi_h - 0.001)
    # construct contour polygon
    point_list = []
    num_point = cnt_arr.shape[1]
    for ind in np.arange(num_point):
        # need to change h-w to w-h
        point_list.append((cnt_arr[1][ind], cnt_arr[0][ind]))
    cnt_poly = geometry.Polygon(point_list)

    in_flag = wsi_poly.contains(cnt_poly)

    return in_flag


def intersect_cnt_wsi(cnt_arr, wsi_h, wsi_w):
    """ Cutting out the contour part inside the whole slide image.

    Parameters
    -------
    cnt_arr: np.array
        contour with standard numpy 2d array format
    wsi_h: int
        height of whole slide image
    wsi_w: int
        width of whole slide image

    Returns
    -------
    inter_cnt: np.array
        contour intersected with whole slide image

    """

    if cnt_inside_wsi(cnt_arr, wsi_h, wsi_w) == True:
        inter_cnt = cnt_arr.astype(np.uint32)
    else:
        # we remove the last line in both width and height of contour
        wsi_poly = geometry.box(0, 0, wsi_w - 1, wsi_h - 1)

        # construct contour polygon
        point_list = []
        num_point = cnt_arr.shape[1]
        for ind in np.arange(num_point):
            # need to change h-w to w-h
            point_list.append((cnt_arr[1][ind], cnt_arr[0][ind]))
        cnt_poly = geometry.Polygon(point_list)

        # get the intersection part of two polygon
        inter_poly = wsi_poly.intersection(cnt_poly)

        x_coors, y_coors = inter_poly.exterior.coords.xy
        x_coors = x_coors[:-1].tolist()
        y_coors = y_coors[:-1].tolist()
        inter_cnt = np.zeros((2, len(x_coors)), dtype=np.uint32)
        for ind in np.arange(len(x_coors)):
            inter_cnt[0, ind] = y_coors[ind]
            inter_cnt[1, ind] = x_coors[ind]

    return inter_cnt


def cnt_inside_ratio(cnt_arr1, cnt_arr2):
    """ Calculate the ratio between intersection part of cnt_arr1 and cnt_arr2
    to cnt_arr1.

    Parameters
    -------
    cnt_arr1: np.array
        contour with standard numpy 2d array format
    cnt_arr2: np.array
        contour with standard numpy 2d array format

    Returns
    -------
    ratio: float
        intersection ratio of cnt_arr1

    """

    # construct contour polygon
    point_list1, point_list2 = [], []
    num_point1 = cnt_arr1.shape[1]
    num_point2 = cnt_arr2.shape[1]
    # need to change h-w to w-h
    for ind in np.arange(num_point1):
        point_list1.append((cnt_arr1[1][ind], cnt_arr1[0][ind]))
    for ind in np.arange(num_point2):
        point_list2.append((cnt_arr2[1][ind], cnt_arr2[0][ind]))
    cnt_poly1 = geometry.Polygon(point_list1)
    cnt_poly1 = cnt_poly1.convex_hull
    cnt_poly2 = geometry.Polygon(point_list2)
    cnt_poly2 = cnt_poly2.convex_hull

    inter_flag = cnt_poly1.intersects(cnt_poly2)
    if inter_flag == False:
        ratio = 0.0
    else:
        inter_poly = cnt_poly1.intersection(cnt_poly2)
        inter_area = inter_poly.area
        cnt1_area = cnt_poly1.area
        ratio = inter_area * 1.0 / cnt1_area

    return ratio



# -*- coding: utf-8 -*-

import os, sys
import numpy as np
import itertools, uuid
from skimage import io, transform
import openslide


__all__ = ["wsi_coor_splitting",
           "wsi_stride_splitting"]


def wsi_coor_splitting(wsi_h, wsi_w, length, overlap_flag=True):
    """ Spltting whole slide image to starting coordinates.

    Parameters
    -------
    wsi_h: int
        height of whole slide image
    wsi_w: int
        width of whole slide image
    length: int
        length of the patch image
    overlap_flag: boolean
        patch with overlap or not

    Returns
    -------
    coors_arr: list
        list of starting coordinates of patches ([0]-h, [1]-w)

    """

    coors_arr = []
    # splitting in both w and h direction with overlap
    def split_patch_overlap(ttl_len, sub_len):
        p_sets = []
        if ttl_len < sub_len:
            return p_sets
        if ttl_len == sub_len:
            p_sets.append(0)
            return p_sets

        p_num = int(np.ceil(ttl_len * 1.0 / sub_len))
        overlap_len = (p_num * sub_len - ttl_len) * 1.0 / (p_num - 1)
        extend_len = sub_len - overlap_len
        for ind in np.arange(p_num):
            p_sets.append(int(round(extend_len * ind)))
        return p_sets

    # splitting in both w and h direction with overlap
    def split_patch_no_overlap(ttl_len, sub_len):
        p_sets = []
        if ttl_len < sub_len:
            return p_sets
        if ttl_len == sub_len:
            p_sets.append(0)
            return p_sets

        p_num = int(np.floor(ttl_len * 1.0 / sub_len))
        p_sets = [ele*sub_len for ele in np.arange(p_num)]
        return p_sets

    if overlap_flag == True:
        h_sets = split_patch_overlap(wsi_h, length)
        w_sets = split_patch_overlap(wsi_w, length)
    else:
        h_sets = split_patch_no_overlap(wsi_h, length)
        w_sets = split_patch_no_overlap(wsi_w, length)

    # combine points in both w and h direction
    if len(w_sets) > 0 and len(h_sets) > 0:
        coors_arr = list(itertools.product(h_sets, w_sets))

    return coors_arr


def wsi_stride_splitting(wsi_h, wsi_w, patch_len, stride_len):
    """ Spltting whole slide image to patches by stride.

    Parameters
    -------
    wsi_h: int
        height of whole slide image
    wsi_w: int
        width of whole slide image
    patch_len: int
        length of the patch image
    stride_len: int
        length of the stride

    Returns
    -------
    coors_arr: list
        list of starting coordinates of patches ([0]-h, [1]-w)

    """

    coors_arr = []
    def stride_split(ttl_len, patch_len, stride_len):
        p_sets = []
        if patch_len > ttl_len:
            raise AssertionError("patch length larger than total length")
        elif patch_len == ttl_len:
            p_sets.append(0)
        else:
            stride_num = int(np.ceil((ttl_len - patch_len) * 1.0 / stride_len))
            for ind in range(stride_num+1):
                cur_pos = int(((ttl_len - patch_len) * 1.0 / stride_num) * ind)
                p_sets.append(cur_pos)

        return p_sets

    h_sets = stride_split(wsi_h, patch_len, stride_len)
    w_sets = stride_split(wsi_w, patch_len, stride_len)

    # combine points in both w and h direction
    if len(w_sets) > 0 and len(h_sets) > 0:
        coors_arr = list(itertools.product(h_sets, w_sets))

    return coors_arr
